﻿using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.SceneManagement;
using UnityEngine.UI;

namespace AudioMate {
    /// <summary>
    /// AudioMate by dub
    /// -------------------------------------------------------
    /// Virt-a-Mate plugin for randomized audio clip playback
    /// and audio clip organization
    /// -------------------------------------------------------
    /// https://github.com/dub-a-licious/audio-mate
    /// </summary>
    public class AudioMateController : MVRScript
    {
        public static string GuessInitialReceivingNode(Atom atom)
        {
            switch (atom.category)
            {
                case "People":
                    return "HeadAudioSource";
                case "Sound":
                    switch (atom.type)
                    {
                        case "AudioSource":
                            return "AudioSource";
                        case "RhythmAudioSource":
                            return "RhythmSource";
                        case "AptSpeaker":
                            return "AptSpeaker_Import";
                        default:
                            return "None";
                    }
                default:
                    return "None";
            }
        }
        public class ContainingAtomNameChangedEventArgs { public string Before; public string After; }
        public class ContainingAtomNameChangedEvent : UnityEvent<ContainingAtomNameChangedEventArgs> {};
        public readonly ContainingAtomNameChangedEvent OnContainingAtomNameChanged = new ContainingAtomNameChangedEvent();

        private Scene _activeScene;

        public CollectionManager collections;
        public AudioMateClipLibrary audioMateClipLibrary;
        public UIManager ui;
        public new MVRPluginManager Manager => base.manager;
        public FileManager fileManager;
        public TriggerManager Triggers;

        public Atom receivingAtom;
        public JSONStorable receivingNode;

        public bool needsProvisionalClipProcessing;
        public bool uiInitialized;

        private bool _initializingUI;

        private string _missingReceiverNodeStoreId = "";

        private JSONStorableString _fileManagerJSON;
        public JSONStorableStringChooser CollectionsJSON;
        private JSONStorableAction _playRandomClipActionJSON;
        private JSONStorableAction _queueRandomClipActionJSON;
        public JSONStorableStringChooser ReceivingAtomJSON;
        public JSONStorableStringChooser ReceivingNodeJSON;
        public JSONStorableStringChooser TriggerColliderChooserJSON;

        public AudioMateClipLibrary.ClipCursor Cursor => audioMateClipLibrary.Cursor;
        public List<AudioMateClip> Clips => audioMateClipLibrary.Clips;

        #region Initialization
        public override void Init()
        {
            Log("### Init Main Controller ###");
            try
            {
                _activeScene = SceneManager.GetActiveScene();
                base.Init();
                Triggers = new TriggerManager(this);
                InitStorables();
                InitFileManager();
                InitClipLibrary();
                StartCoroutine(DeferredInit());
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(Init)}: {e}");
            }
        }

        protected override void Awake()
        {
            SuperController.singleton.onSceneLoadedHandlers += OnSceneLoaded;
            base.Awake();
        }

        private void OnSceneLoaded()
        {
            if (this != null && collections != null && collections) containingAtom.RestoreFromLast(this);
            Log("--- OnSceneLoaded ---");
        }

        private IEnumerator DeferredInit()
        {
            if (collections == null)
            {
                while (_activeScene.isLoaded == false || SuperController.singleton.isLoading)
                {
                    yield return null;
                    if (this == null) yield break;
                }

                while (!uiInitialized)
                {
                    yield return new WaitForSeconds(.1f);
                    if (this == null) yield break;
                }

                InitCollections();
            }

            Log("Trying to restore from last save...");
            containingAtom.RestoreFromLast(this);
        }

        public override void InitUI()
        {
            Log("### InitActiveCollectionUI ###");

            base.InitUI();
            try
            {
                if (UITransform == null) return;

                SuperController.singleton.StartCoroutine(InitUIDeferred());
            }
            catch (Exception exc)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitUI)}: {exc}");
            }
        }

        private IEnumerator InitUIDeferred()
        {
            if (ui != null ||
                _initializingUI) yield break;
            _initializingUI = true;
            yield return SuperController.singleton.StartCoroutine(UIFactory.LoadUIAssets());

            var scriptUI = UITransform.GetComponentInChildren<MVRScriptUI>();
            var scrollRect = scriptUI.fullWidthUIContent.transform.parent.parent.parent.GetComponent<ScrollRect>();

            if (scrollRect == null)
                SuperController.LogError(
                    "AudioMate: Error during UI initialization. Scroll rect not at the expected hierarchy position");
            else
            {
                scrollRect.vertical = false;
                scrollRect.horizontal = false;
                scrollRect.elasticity = 0;
                scrollRect.inertia = false;
                scrollRect.movementType = ScrollRect.MovementType.Clamped;
            }

            ui = UIManager.AddTo(scriptUI.fullWidthUIContent);
            ui.Init(this);
            var count = 0;
            while (collections == null || !collections.isInitialized)
            {
                // security break out to avoid endless loops
                if (++count > 200)
                {
                    SuperController.LogError(!collections.isInitialized
                        ? "AudioMate: Could not initialize collection manager."
                        : "AudioMate: Collection manager is null.");

                    yield break;
                }
                yield return new WaitForSeconds(.1f);
                if (this == null) yield break;
            }
            try
            {
                ui.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate: {e}");
            }
        }

        public void OnUIInitialized()
        {
            Log("### OnUIInitialized ###");
            uiInitialized = true;
        }

        private void InitCollections()
        {
            if ((UnityEngine.Object) collections != (UnityEngine.Object) null) return;
            Log("### InitCollections ###");
            try
            {
                collections = gameObject.AddComponent<CollectionManager>();
                collections.transform.SetParent(transform, false);

                collections.OnActiveCollectionSelected.RemoveAllListeners();
                collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
                collections.OnActiveCollectionNameChanged.RemoveAllListeners();
                collections.OnActiveCollectionNameChanged.AddListener(OnActiveCollectionNameChanged);
                _playRandomClipActionJSON = new JSONStorableAction(Storables.PlayRandomClipAction,
                    () => collections.PlayRandomClipAction());
                RegisterAction(_playRandomClipActionJSON);
                _queueRandomClipActionJSON = new JSONStorableAction(Storables.QueueRandomClipAction,
                    () => collections.QueueRandomClipAction());
                RegisterAction(_queueRandomClipActionJSON);
                collections.Init(this);
                audioMateClipLibrary.Bind();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(InitCollections)}: {e}");
            }
        }

        private void InitFileManager()
        {
            Log("### InitFileManager ###");
            fileManager = gameObject.AddComponent<FileManager>();
            fileManager.Bind(_fileManagerJSON);
        }

        private void InitClipLibrary()
        {
            Log("### InitClipLibrary ###");
            audioMateClipLibrary = gameObject.AddComponent<AudioMateClipLibrary>();
            audioMateClipLibrary.Init(this);
        }

        private void InitStorables()
        {
            Log("### InitStorables ###");

            CollectionsJSON = new JSONStorableStringChooser(
                Storables.ClipCollections,
                new List<string>(),
                "",
                "Active Collection")
            {
                isRestorable = false,
                isStorable = false
            };
            RegisterStringChooser(CollectionsJSON);

            ReceivingAtomJSON = new JSONStorableStringChooser(Storables.ReceivingAtom, null,
                containingAtom.uid, "Receiver Atom", SetReceivingAtom)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            RegisterStringChooser(ReceivingAtomJSON);

            ReceivingNodeJSON = new JSONStorableStringChooser(Storables.ReceivingNode, null,
                GuessInitialReceivingNode(containingAtom), "Receiver Node", SetReceivingNode)
            {
                storeType = JSONStorableParam.StoreType.Full
            };
            RegisterStringChooser(ReceivingNodeJSON);

            TriggerColliderChooserJSON = new JSONStorableStringChooser(
                Storables.TriggerActions,
                Triggers.ColliderChoices,
                TriggerManager.DefaultCollider,
                "Trigger");
            RegisterStringChooser(TriggerColliderChooserJSON);

            _fileManagerJSON = new JSONStorableString("File Manager", "")
            {
                isStorable = false,
                isRestorable = false
            };
            RegisterString(_fileManagerJSON);
        }
        #endregion

        #region Lifecycle
        public void OnEnable()
        {
            if (collections == null) return;
            collections.EnableAll();
        }

        public void OnDisable()
        {
            if (collections != null) collections.DisableAll();
        }

        public void Start()
        {
            Log("### Start ###");
            //var material = ui.container.GetComponent<Renderer>().material;
            //material.color = Color.clear;
           //StartCoroutine(Tools.LerpMaterialColor(material, 1,
            //Color.clear, new Color(1, 1, 1, 1)));
        }

        public void OnDestroy()
        {
            Log("### OnDestroy ###");
            Triggers.OnDestroy();
            if (ui != null) Destroy(ui);
            if (collections != null) Destroy(collections);
        }

        public void Update()
        {
            if (!uiInitialized || collections == null) return;
            if (audioMateClipLibrary != null && needsProvisionalClipProcessing && audioMateClipLibrary.isInitialised)
            {
                needsProvisionalClipProcessing = false;
                audioMateClipLibrary.ProcessProvisionalClips();
            }
            FindMissingReceiverNode();
        }

        /**
		 * Create/refresh the list of available custom scene clips.
		 */
        public void RefreshClipLibrary()
        {
            if ((UnityEngine.Object) audioMateClipLibrary == (UnityEngine.Object) null) return;
            audioMateClipLibrary.RefreshClipIndex();
        }

        protected void OnAtomRename(string oldID, string newID)
        {
            if ((UnityEngine.Object) collections == (UnityEngine.Object) null) return;
            collections.OnAtomRename(oldID, newID);
            Triggers?.OnAtomRename(oldID, newID);
        }
        #endregion

        #region Load / Save
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            var json = base.GetJSON(includePhysical, includeAppearance, forceStore);

            if (collections == null) return json;
            collections.Serialize(json, Storables.ClipCollections);
            needsStore = true;

            return json;
        }
        private bool _skipBaseRestore;
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            if (_skipBaseRestore == false)
            {
                base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            }
            else
            {
                Log(" - Late Restore");
                _skipBaseRestore = false;
            }
            try
            {
                if (jc == null || jc.AsObject.HasKey(Storables.ClipCollections) == false)
                {
                    Log("RestoreFromJSON: with empty json object: {jc}. Aborting restore!");
                    return;
                }
                var restoredJSON = jc.AsObject[Storables.ClipCollections];
                if (restoredJSON == null)
                {
                    Log("RestoreFromJSON: with empty json object: {jc}. Aborting restore!");
                    return;
                }
                SuperController.singleton.StartCoroutine(LoadDeferred(restoredJSON));
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(RestoreFromJSON)}: {e}");
            }
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true,
            bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            _skipBaseRestore = true;
            RestoreFromJSON(jc, restorePhysical, restoreAppearance, null, setMissingToDefault);
            _skipBaseRestore = true;
        }

        public void Load(JSONNode json)
        {
            if (collections == null) return;
            try
            {
                if (json == null || json.AsObject == null || $"{json}" == "{}")
                {
                    Log("Load called with empty json object: {json}. Aborting load!");
                    return;
                }
                collections.Parse(json);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(Load)}: {e}");
            }
            finally
            {
                insideRestore = false;
            }
        }

        private IEnumerator LoadDeferred(JSONNode json)
        {
            if (insideRestore) yield break;
            insideRestore = true;
            while (collections == null)
            {
                yield return null;
                if (this == null) yield break;
            }
            Load(json);
        }
        #endregion

        #region File Import
        private string ImportAudioFiles(string requestedPath)
        {
            if (_fileManagerJSON == null) return null;
            try
            {
                SuperController.singleton.GetFilesAtPath(requestedPath).ToList().ForEach((string fileName) =>
                {
                    if (fileName.Contains(".json"))
                    {
                        return;
                    }

                    if (!fileName.Contains(".mp3") && !fileName.Contains(".wav") && !fileName.Contains(".ogg"))
                    {
                        return;
                    }

                    LoadAudio(fileName);
                    _fileManagerJSON.val += SuperController.singleton.NormalizePath(fileName) + "\n";
                });
                audioMateClipLibrary.OnSourceClipsUpdated();
                return requestedPath;
            }
            catch
            {
                var folderName = "\\" + requestedPath.Substring(requestedPath.LastIndexOf('\\') + 1) + "\\";
                requestedPath = requestedPath.Replace(folderName, "\\");
                SuperController.singleton.GetFilesAtPath(requestedPath).ToList().ForEach((string fileName) =>
                {
                    if (fileName.Contains(".json"))
                    {
                        return;
                    }

                    if (!fileName.Contains(".mp3") && !fileName.Contains(".wav") && !fileName.Contains(".ogg"))
                    {
                        return;
                    }

                    //clipsSource.Add(LoadAudio(fileName));
                    LoadAudio(fileName);
                    _fileManagerJSON.val += SuperController.singleton.NormalizePath(fileName) + "\n";
                });
                audioMateClipLibrary.OnSourceClipsUpdated();
                return requestedPath;
            }
        }

        private static NamedAudioClip LoadAudio(string path)
        {
            var localPath = SuperController.singleton.NormalizeLoadPath(path);
            var existing = URLAudioClipManager.singleton.GetClip(localPath);
            if (existing != null)
            {
                return existing;
            }

            var clip = URLAudioClipManager.singleton.QueueClip(SuperController.singleton.NormalizeMediaPath(path));
            if (clip == null)
            {
                return null;
            }

            var nac = URLAudioClipManager.singleton.GetClip(clip.uid);
            return nac ?? null;
        }

        #endregion

        #region Helpers

        public AudioMateClip FindAudioMateClipBySourceClipUID(string sourceClipUID)
        {
            var audioMateClip = audioMateClipLibrary.FindBySourceClipUID(sourceClipUID);
            return audioMateClip;
        }
        #endregion
        
        #region Callbacks

        public void OnAddTriggerButtonClicked(string triggerType = TriggerManager.StartTriggerActions)
        {
            if (collections == null) return;
            if (collections.ActiveCollection == null) return;
            var triggerName = GetPlayRandomClipTriggerName(TriggerColliderChooserJSON.val, collections.ActiveCollection.Name);
            if (string.IsNullOrEmpty(triggerName)) return;
            var receiverTargetName = CollectionManager.GetPlayRandomClipActionName(collections.ActiveCollection.Name);
            Triggers.AddTriggerAction(TriggerColliderChooserJSON.val, triggerName, storeId, receiverTargetName, triggerType);
        }

        private void UpdateTriggerActionNamesOnCollectionRename(string oldCollectionName, string newCollectionName)
        {
            if (string.IsNullOrEmpty(Tools.TrimAll(oldCollectionName)) || string.IsNullOrEmpty(Tools.TrimAll(newCollectionName))) return;

            foreach (var triggerCollider in TriggerColliderChooserJSON.choices)
            {
                var oldTriggerActionName = GetPlayRandomClipTriggerName(triggerCollider, oldCollectionName);
                var newTriggerActionName = GetPlayRandomClipTriggerName(triggerCollider, newCollectionName);
                var newTriggerActionReceiverName = CollectionManager.GetPlayRandomClipActionName(newCollectionName);

                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName, newTriggerActionName, newTriggerActionReceiverName);
                Triggers.UpdateTriggerAction(triggerCollider, oldTriggerActionName, newTriggerActionName, newTriggerActionReceiverName, TriggerManager.EndTriggerActions);
            }
        }

        private static string GetPlayRandomClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetPlayRandomClipActionName(collectionName)}");
        }

        private  string GetQueueRandomClipTriggerName(string triggerCollider, string collectionName)
        {
            return Tools.TrimAll($"{triggerCollider}:{CollectionManager.GetQueueRandomClipActionName(collectionName)}");
        }

        public void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs eventArgs)
        {
            if (collections == null) return;

            var atom = collections.ActiveCollection?.ReceiverAtom;
            if (atom == null) return;
            ReceivingAtomJSON.val = atom;

            var node = collections.ActiveCollection?.ReceiverNode;
            if (node == null) return;
            ReceivingNodeJSON.val = node;
        }

        public void OnActiveCollectionNameChanged(CollectionManager.ActiveCollectionNameChangedEventArgs eventArgs)
        {
            UpdateTriggerActionNamesOnCollectionRename(eventArgs.Before, eventArgs.After);
        }

        /**
		* Load audio source ReceiverAtom.
		*/
        private void SetReceivingAtom(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            try
            {
                receivingAtom = SuperController.singleton.GetAtomByUid(uid);
                if (receivingAtom == null)
                {
                    SuperController.LogError(
                        $"AudioMate.{nameof(AudioMateController)}.{nameof(SetReceivingAtom)}: Audio Source ReceiverAtom has invalid uid {uid}");
                    return;
                }
                var atomChanged = collections.ActiveCollection.ReceiverAtom != receivingAtom.uid;
                collections.ActiveCollection.ReceiverAtom = receivingAtom.uid;
                if (!atomChanged) return;
                var guessedReceiver = GuessInitialReceivingNode(receivingAtom);
                insideRestore = true;
                ReceivingNodeJSON.val = guessedReceiver;
                insideRestore = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(SetReceivingAtom)}: {e}");
            }
        }

        /**
		* Load audio source receiver.
		*/
        private void SetReceivingNode(string uid)
        {
            if (Tools.IsEmptyChoice(uid) || (UnityEngine.Object) collections == (UnityEngine.Object) null || collections.ActiveCollection == null) return;
            receivingNode = SuperController.singleton.GetAtomByUid(ReceivingAtomJSON.val)?.GetStorableByID(uid);
            if (receivingNode == null)
            {
                _missingReceiverNodeStoreId = uid;
                return;
            }

            insideRestore = true;
            ReceivingNodeJSON.val = receivingNode.name;
            insideRestore = false;
            collections.ActiveCollection.ReceiverNode = receivingNode.name;
            collections.ActiveCollection.SyncAudioReceiver();
        }

        private void FindMissingReceiverNode() {
            if (_missingReceiverNodeStoreId != "" && receivingAtom != null) {
                var missingReceiver = receivingAtom.GetStorableByID(_missingReceiverNodeStoreId);
                if (missingReceiver == null) return;
                SetReceivingNode(_missingReceiverNodeStoreId);
                _missingReceiverNodeStoreId = "";
            }
        }
        #endregion

        #region Debug
        /**
         * Custom debug logging
         */
        public void Log(string message)
        {
            #if DEBUG
            SuperController.LogMessage($"[{this.storeId}]: {message}");
            #endif
        }

        /**
         * UI debugging. Depends on VAM Dev Tools by Acidbubbles (https://github.com/acidbubbles/vam-devtools).
         */
        public void DebugObject(Transform obj)
        {
            #if DEBUG
            DebugObject(obj.gameObject);
            #endif
        }

        public void DebugObject(GameObject obj)
        {
            #if DEBUG
            containingAtom.BroadcastMessage("DevToolsInspectUI", obj);
            #endif
        }
        #endregion
    }
}
