﻿using System;
using MVR.FileManagement;
using SimpleJSON;
using UnityEngine;
using UnityEngine.UI;
using Object = UnityEngine.Object;

namespace AudioMate
{
    public class AudioMateClipUI
    {
        public bool IsInitialised;
        public Transform Container;
        public UIDynamicButton PreviewButton;
        public UIDynamicButton ToggleButton;
        public Outline ToggleOutline;

        public bool SetInCollectionState(bool state)
        {
            if (ToggleButton == null) return state;
            ToggleButton.textColor = state ? Styles.EnabledText : Styles.DisabledText;
            ToggleButton.buttonColor = state ? Styles.EnabledBg : Styles.DisabledBg;
            if (ToggleOutline != null) ToggleOutline.effectColor = state ? Color.white : Styles.DisabledBg;

            return state;
        }

        public bool SetCursorState(bool state)
        {
            if (PreviewButton == null) return state;
            PreviewButton.textColor = state ? Styles.CursorEnabledText : Styles.CursorDisabledText;
            PreviewButton.buttonColor = state ? Styles.CursorEnabledBg : Styles.CursorDisabledBg;
            return state;
        }

        public void Destroy()
        {
            try
            {
                if (Container != null) Object.Destroy(Container.gameObject);
                Container = null;
                if (PreviewButton != null) Object.Destroy(PreviewButton.gameObject);
                PreviewButton = null;
                if (ToggleButton != null) Object.Destroy(ToggleButton.gameObject);
                ToggleButton = null;
                if (ToggleOutline != null) Object.Destroy(ToggleOutline.gameObject);
                ToggleOutline = null;
                IsInitialised = false;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipUI)}.{nameof(Destroy)}: {e}");
            }
        }
    }

    public class AudioMateClip
    {
        public string DisplayName => SourceClip.displayName;

        public string UID
        {
            get {
                if (_uid == null && SourceClip != null)
                {
                    _uid = $"AudioMateClip{SourceClip.uid}";
                }

                return _uid;
            }
        }

        private string _uid;
        public NamedAudioClip SourceClip { get; private set; }
        public AudioSourceControl Receiver;

        public bool IsInActiveCollection { get; private set; }
        public bool HasCursor { get; private set; }

        public AudioMateClipUI UI;

        public bool IsProvisional;

        public string SourcePath;

        public AudioMateClip(NamedAudioClip sourceClip)
        {
            SourceClip = sourceClip;
            Init();
        }

        public AudioMateClip(JSONNode jn)
        {
            FromJSON(jn);
            Init();
        }

        private void Init()
        {
            UI = new AudioMateClipUI();
            IsInActiveCollection = false;
            HasCursor = false;
            AddSourcePath();
        }

        public void RefreshUI()
        {
            if (UI == null || UI.IsInitialised == false) return;
            UI.SetInCollectionState(IsInActiveCollection);
            UI.SetCursorState(HasCursor);
        }

        public JSONClass ToJSON()
        {
            return new JSONClass
            {
                { "uid", UID },
                { "sourceClip", SourceClip?.uid },
                { "hasCursor", HasCursor.ToString() },
                { "sourcePath", SourcePath }
            };
        }

        private void FromJSON(JSONNode jn)
        {
            if (jn == null || jn.AsObject == null) return;
            var clipUID = jn["sourceClip"] != null ? jn["sourceClip"] : jn["sourceClipUID"];
            SourcePath = jn["sourcePath"];
            SourceClip = URLAudioClipManager.singleton.GetClip(clipUID);
            if (SourceClip == null && SourcePath != null)
            {
                URLAudioClipManager.singleton.QueueClip(SourcePath);
            }
            if (string.IsNullOrEmpty(SourcePath))
            {
                AddSourcePath();
            }
        }

        private void AddSourcePath()
        {
            if (SourceClip == null) return;
            var sceneClipData = URLAudioClipManager.singleton.GetJSON()["clips"];
            foreach (var clipData in sceneClipData.Childs)
            {
                if (clipData["displayName"].Value == SourceClip.displayName)
                {
                    SourcePath = clipData["url"];
                }
            }
        }

        /**
         * Play the assigned audio clip with optional "if clear" param.
         */
        public void Play(AudioSourceControl receiver = null, bool ifClear = false, bool clearQueue = false)
        {
            if (SourceClip == null) return;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null) Receiver = receiver;
            if (ifClear) {
                Receiver.PlayIfClear(SourceClip);
            } else {
                if (clearQueue)
                {
                    Receiver.PlayNowClearQueue(SourceClip);
                }
                else
                {
                    Receiver.PlayNow(SourceClip);
                }
            }
        }

        /**
         * Add assigned audio clip to audio _c queue list.
         */
        public void Queue(AudioSourceControl receiver = null)
        {
            if (SourceClip == null) return;
            if ((UnityEngine.Object) receiver != (UnityEngine.Object) null)
            {
                Receiver = receiver;
            }
            else
            {
                return;
            }
            Receiver.QueueClip(SourceClip);
        }

        public void Destroy()
        {
            try
            {
                UI.Destroy();
                Receiver = null;
                SourceClip = null;
            } catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClip)}.{nameof(Destroy)}: {e}");
            }
        }

        public bool ToggleState()
        {
            IsInActiveCollection = !IsInActiveCollection;
            UI.SetInCollectionState(IsInActiveCollection);
            return IsInActiveCollection;
        }

        public bool SetState(bool state)
        {
            IsInActiveCollection = state;
            UI.SetInCollectionState(state);
            return state;
        }

        public bool SetCursor(bool state)
        {
            HasCursor = state;
            UI.SetCursorState(state);
            return state;
        }
    }
}
