﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public class AudioMateClipCollection : IEnumerable<AudioMateClip>
    {
        public class ClipPlayedEvent : UnityEvent {}
        public readonly ClipPlayedEvent OnClipPlayed = new ClipPlayedEvent();

        public string Name;
        public bool Enabled = true;
        public bool ShuffleMode = true;
        public bool AlwaysQueue;
        public bool OnlyIfClear;
        public string ReceiverAtom;
        public string ReceiverNode;
        public AudioSourceControl Receiver;

        public float PlayChance = 1f;

        private int _lastPlayedClipIndex = -1;

        private readonly List<int> _unplayedClipsList = new List<int>();

        public IEnumerable<NamedAudioClip> NamedAudioClips
        {
            get
            {
                return Enumerable.ToList(Clips.Select(clip => clip.SourceClip));
            }
        }

        public List<AudioMateClip> Clips { get; } = new List<AudioMateClip>();

        public int Count => Clips.Count;

        public bool Contains(AudioMateClip clip) => (bool) Clips.Contains(clip);
        public AudioMateClip Find(System.Predicate<AudioMateClip> predicate) => Clips.Find(predicate);

        public AudioMateClip LastPlayedClip
        {
           get {
                if (_lastPlayedClipIndex == -1) return null;
                return Clips.Count > _lastPlayedClipIndex ? Clips[_lastPlayedClipIndex] : null;
           }
        }


        public AudioMateClipCollection(AudioMateController controller)
        {
            ResetUnplayedClips();
            Init();
        }

        public AudioMateClipCollection(string collectionName)
        {
            Name = collectionName;
            ResetUnplayedClips();
            Init();
        }

        public AudioMateClipCollection(string collectionName, IEnumerable<AudioMateClip> clips)
        {
            try
            {
                Name = collectionName;
                Clips.AddRange(clips);
                ResetUnplayedClips();
                Init();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(AudioMateClipCollection)} (IEnumerable): {e}");
            }
        }

        public AudioMateClipCollection(JSONNode jn)
        {
            if (jn == null || !jn.Childs.Any()) return;
            try
            {
                Parse(jn);
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(AudioMateClipCollection)} (JSON): {e}");
            }
        }

        private void Init()
        {
            SetDefaults();
        }

        private void SetDefaults()
        {
            SyncAudioReceiver();
        }

        public JSONClass ToJSON()
        {
            var clipsJSON = new JSONArray();
            if (Clips.Count > 0)
            {
                foreach (var clip in Clips)
                {

                    clipsJSON.Add(clip.ToJSON());
                }
            }

            var collectionJSON = new JSONClass
            {
                { "name", Name },
                { "enabled", Enabled.ToString() },
                { "receiverAtom", ReceiverAtom },
                { "receiverNode", ReceiverNode },
                { "shuffle", ShuffleMode.ToString() },
                { "alwaysQueue", AlwaysQueue.ToString() },
                { "onlyIfClear", OnlyIfClear.ToString() },
                { "playChance", PlayChance.ToString(CultureInfo.InvariantCulture) },
                { "lastClipIndex", _lastPlayedClipIndex.ToString() },
                { "clips", clipsJSON }
            };
            return collectionJSON;
        }

        public bool Parse(JSONNode jn, bool merge = false)
        {
            try
            {
                if (jn == null || jn.AsObject == null)
                {
                    SuperController.LogMessage($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(Parse)} param jn is null");
                    return false;
                }

                if (!merge) Clips.Clear();
                Name = jn["name"];
                Enabled = jn["enabled"].AsBool;
                ReceiverAtom = jn["receiverAtom"];
                ReceiverNode = jn["receiverNode"];
                ShuffleMode = jn["shuffle"].AsBool;
                AlwaysQueue = jn["alwaysQueue"].AsBool;
                OnlyIfClear = jn["onlyIfClear"].AsBool;
                PlayChance = jn["playChance"].AsFloat;
                _lastPlayedClipIndex = merge ? -1 : jn["lastClipIndex"].AsInt;
                if (jn["clips"] != null)
                {
                    foreach (JSONClass clip in jn["clips"].AsArray)
                    {
                        if (HasClip(clip)) continue;
                        var provisionalClip = new AudioMateClip(clip) { IsProvisional = true };
                        Clips.Add(provisionalClip);
                    }
                }
                ResetUnplayedClips();
                Init();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(Parse)} {e}");
                return false;
            }
            return true;
        }

        /**
         * Adds a clip to this collections
         */
        public void Add(AudioMateClip clip)
        {
            if (Clips == null) return;
            if (Clips.Contains(clip)) return;
            Clips.Add(clip);
            ResetUnplayedClips();
        }

        /**
         * Removes a clip from this collection
         */
        public bool Remove(NamedAudioClip sourceClip)
        {
            if (Clips == null) return false;
            var clip = FindByNamedAudioClip(sourceClip);
            if (clip == null) return false;
            ResetUnplayedClips();
            return true;
        }

        /**
         * Removes a clip from this collection
         */
        public bool Remove(AudioMateClip clip)
        {
            if (Clips == null) return false;
            var result = Clips.Remove(clip);
            if (result) ResetUnplayedClips();
            return result;
        }

        public void Clear()
        {
            foreach (var clip in Clips)
            {
                clip.Destroy();
            }
            Clips.Clear();
            ResetUnplayedClips();
        }

        public bool HasClip(AudioMateClip clip)
        {
            return HasClip(clip.ToJSON());
        }

        public bool HasClip(JSONClass clip)
        {
            return Clips.Exists(x => clip["sourceClip"] == x.SourceClip);
        }

        public void SwapProvisionalClip(AudioMateClip provisionalClip, AudioMateClip audioMateClip)
        {
            var index = Clips.IndexOf(provisionalClip);
            Clips.Remove(provisionalClip);
            Clips.Insert(index, audioMateClip);
        }

        public void Enable()
        {
            Enabled = true;
        }

        public void Disable()
        {
            Enabled = false;
        }

        public List<AudioMateClip> GetProvisionalClips()
        {
            return Clips.FindAll(x => x.IsProvisional);
        }

        /**
         * Checks if a NamedAudioClip already exists in this collection.
         */
        public bool ContainsNamedAudioClip(NamedAudioClip namedAudioClip)
        {
            return ContainsNamedAudioClip(namedAudioClip.uid);
        }

        private bool ContainsNamedAudioClip(string namedAudioClipUID)
        {
            return Clips != null && Clips.Exists(x => x.SourceClip.uid == namedAudioClipUID);
        }

        private AudioMateClip FindByNamedAudioClip(NamedAudioClip sourceClip)
        {
            return Clips?.Find(x => x.SourceClip == sourceClip);
        }

        /**
         * Evaluate play chance and if it succeeds get a randomly (shuffle) selected clip from the clip collection.
         */
        private AudioMateClip GetRandomClip(bool skipPlayChance = false)
        {
            if (Clips == null) return null;
            var randomClip = (AudioMateClip) null;
            var randomIndex = -1;

            try
            {
                // PlayChance test
                var randomChance = (float) UnityEngine.Random.Range(0, 100) / 100;

                if (!skipPlayChance && PlayChance < randomChance) return null;

                if (ShuffleMode)
                {
                    if (_unplayedClipsList.Count == 1)
                    {
                        randomClip = Clips[_unplayedClipsList.First()];

                        _lastPlayedClipIndex = _unplayedClipsList.ElementAtOrDefault(0);
                        ResetUnplayedClips();
                        return randomClip;
                    }
                    // ShuffleMode play
                    if (_unplayedClipsList.Count < 1)
                    {
                        ResetUnplayedClips();
                    }

                    // After playing the last clip of the unplayed clips list and resetting it can theoretically happen that
                    // the next random clip is the same as the last random clip. If that happens lets repeat randomization
                    // until we get another clip index.
                    randomIndex = UnityEngine.Random.Range(0, _unplayedClipsList.Count);
                    if (Clips.Count > 1)
                    {
                        var safety = 0;
                        while (randomIndex == _lastPlayedClipIndex)
                        {
                            randomIndex = UnityEngine.Random.Range(0, _unplayedClipsList.Count);
                            if (++safety < 100) continue;
                            return null;
                        }
                    }
                    randomClip = Clips.ElementAtOrDefault(_unplayedClipsList.ElementAtOrDefault(randomIndex));

                    // Remove played clip from unplayedClipsList list
                    _unplayedClipsList.RemoveAt(randomIndex);
                }
                else
                {
                    // Standard play
                    randomIndex = UnityEngine.Random.Range(0, Clips.Count);
                    randomClip = Clips.ElementAtOrDefault(randomIndex);
                }

                // Set last clip index to keep track of the current position
                _lastPlayedClipIndex = randomIndex;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(GetRandomClip)}: randomIndex: {randomIndex} {e}");
            }

            return randomClip;
        }

        /**
         * Set the audio source _c which plays the clips
         */
        public void SyncAudioReceiver()
        {
            if (ReceiverAtom == null || ReceiverNode == null) return;
            try
            {
                Receiver =
                    SuperController.singleton.GetAtomByUid(ReceiverAtom)?.GetStorableByID(ReceiverNode)
                        as AudioSourceControl;
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateController)}.{nameof(SyncAudioReceiver)}: {e}");
            }
        }

        /**
		* Play or queue a randomly selected clip from the collection.
		*/
        public void PlayRandomClip() {
            try
            {
                if (!Enabled) return;
                if (Clips.Count == 0 || Receiver == null) return;

                var nextClip = GetRandomClip();
                if (AlwaysQueue)
                {
                    nextClip?.Queue(Receiver);
                } else {
                    nextClip?.Play(Receiver, OnlyIfClear);
                }
                OnClipPlayed.Invoke();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(PlayRandomClip)}: {e}");
            }
        }

        public void QueueRandomClip() {
            try {
                if (!Enabled) return;
                if (Clips.Count < 1 || Receiver == null) return;

                var nextClip = GetRandomClip();
                nextClip?.Queue(Receiver);
                OnClipPlayed.Invoke();
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipCollection)}.{nameof(QueueRandomClip)}: {e}");
            }
        }

        /**
         * Reset the unplayed clips list, which is used for the shuffle mode. The list is a simple integer index list
         * which gets updated after every play of a clip (clip index gets removed).
         */
        private void ResetUnplayedClips()
        {
            _unplayedClipsList.Clear();
            if (Clips == null) return;
            for (var i = 0; i < Clips.Count; i++)
            {
                _unplayedClipsList.Add(i);
            }
        }

        public override string ToString()
        {
            return Clips.Aggregate("", (current, clip) => current + (clip.DisplayName + Environment.NewLine));
        }

        IEnumerator<AudioMateClip> IEnumerable<AudioMateClip>.GetEnumerator()
        {
            return Clips.GetEnumerator();
        }

        public IEnumerator GetEnumerator()
        {
            return Clips.GetEnumerator();
        }
    }
}
