﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{

    public class AudioMateClipLibrary : MonoBehaviour
    {
        public class ClipCursor
        {
            public int Index;
            public AudioMateClip Clip;
        }

        public class ClipsRefreshedEvent : UnityEvent { }
        public readonly ClipsRefreshedEvent OnClipsRefreshed = new ClipsRefreshedEvent();

        private AudioMateController _c;

        private List<NamedAudioClip> _sourceClipList;
        private Coroutine _processing;

        public List<AudioMateClip> Clips { get; private set; }
        public readonly ClipCursor Cursor = new ClipCursor();
        public bool isInitialised;
        private bool _isBound;

        public void Init(AudioMateController controller)
        {
            if (controller == null) return;
            _c = controller;
            Log(" - Init AudioMateClipLibrary");
            Clips = new List<AudioMateClip>();
            _c.fileManager.OnNewFilesImported.AddListener(OnSourceClipsUpdated);
        }

        public void Bind()
        {
            Log(" - Bind");
            _c.collections.OnActiveCollectionSelected.AddListener(OnActiveCollectionSelected);
            _c.collections.OnActiveCollectionUpdated.AddListener(OnActiveCollectionUpdated);
            _c.fileManager.OnNewFilesImported.AddListener(OnSourceClipsUpdated);
            _isBound = true;
            RefreshClipIndex();
        }

        public void SetCursorIndex(int index)
        {
            if (Cursor == null) return;
            Cursor.Index = index;
            Cursor.Clip = Clips[index];
        }

        public void RefreshClipIndex()
        {
            if (_isBound == false || _processing != null) return;
            Log(" - RefreshClipIndex");
            try
            {
                if (_processing == null) _processing = SuperController.singleton.StartCoroutine(DeferredIndexing());
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(AudioMateClipLibrary)}.{nameof(RefreshClipIndex)}: {e}");
            }
        }

        private IEnumerator DeferredIndexing()
        {
            // endless loop counter
            var elc = 0;
            _sourceClipList = URLAudioClipManager.singleton.GetCategoryClips("web");
            while (_sourceClipList == null)
            {
                _sourceClipList = URLAudioClipManager.singleton.GetCategoryClips("web");
                yield return new WaitForSeconds(.1f);
                if (this == null) yield break;
                elc++;
                if (elc <= 1000) continue;
                SuperController.LogError("AudioMate.AudioMateClipLibrary.DeferredIndexing: Endless loop detected!");
                yield break;
            }
            // Convert new clips to AudioMateClips and add them to the list
            var clipCount = 0;
            foreach (var audioMateClip in
                from sourceClip in _sourceClipList
                where Clips.Find(x => x.SourceClip.uid == sourceClip.uid) == null
                select new AudioMateClip(sourceClip))
            {
                Clips.Add(audioMateClip);
                clipCount++;
            }

            Log($" - Added {clipCount} clips to library.");
            var removedClips = RemoveOrphanedClips();
            isInitialised = true;
            if (clipCount > 0 || removedClips > 0) OnClipsRefreshed.Invoke();
            _processing = null;
        }

        // Removes AudioMateClips with null value source clips.
        private int RemoveOrphanedClips()
        {
            var clipCount = 0;
            foreach (var audioMateClip in Clips.Where(audioMateClip => audioMateClip.SourceClip == null))
            {
                Clips.Remove(audioMateClip);
                audioMateClip.Destroy();
                clipCount++;
            }
            Log($" - Removed {clipCount} orphaned clips from the library.");
            return clipCount;
        }

        public AudioMateClip FindBySourceClipUID(string sourceClipUID)
        {
            return Clips.Find(x => x.SourceClip.uid == sourceClipUID);
        }

        public void ProcessProvisionalClips()
        {
            Log($" - Processing provisional clips");

            foreach (var collection in _c.collections.CollectionsWithProvisionalClips)
            {
                Log($" - - Processing collection {collection.Name}.");
                var isActive = collection == _c.collections.ActiveCollection;
                foreach (var provisionalClip in collection.GetProvisionalClips())
                {
                    var audioMateClip = FindBySourceClipUID(provisionalClip.SourceClip.uid);
                    if (audioMateClip == null) continue;
                    if (isActive) audioMateClip.SetState(true);
                    collection.SwapProvisionalClip(provisionalClip, audioMateClip);
                }
            }
            OnClipsRefreshed.Invoke();
        }

        public void OnActiveCollectionSelected(
            CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
        }

        public void OnActiveCollectionUpdated()
        {
        }

        public void OnSourceClipsUpdated()
        {
            Log("Scene Clips updated!");
            RefreshClipIndex();
        }

        public void OnDestroy()
        {
            _isBound = false;
            try
            {
                if (_processing != null) SuperController.singleton.StopCoroutine(_processing);
                _processing = null;
            }
            catch (Exception e)
            {
                SuperController.LogMessage($"AudioMate: [UIClipLibrary.OnDestroy] {e}");
            }
        }

        private void Log(string message)
        {
            if (_c == null) return;
            _c.Log($" [AudioMateClipLibrary]: {message}");
        }
    }
}
