﻿using System;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;
using UnityEngine.Events;

namespace AudioMate
{
    public class CollectionManager : MonoBehaviour
    {

        private class CollectionActions
        {
            public JSONStorableAction PlayAction;
            public JSONStorableAction QueueAction;
        }

        public class ActiveCollectionSelectedEventArgs { public AudioMateClipCollection Before; public AudioMateClipCollection After; }
        public class ActiveCollectionNameChangedEventArgs { public string Before; public string After; }

        public class ActiveCollectionSelectedEvent : UnityEvent<ActiveCollectionSelectedEventArgs> { }
        public class ActiveCollectionUpdatedEvent : UnityEvent { }
        public class ActiveCollectionNameChangedEvent : UnityEvent<ActiveCollectionNameChangedEventArgs> { }

        public ActiveCollectionSelectedEvent OnActiveCollectionSelected = new ActiveCollectionSelectedEvent();
        public ActiveCollectionUpdatedEvent OnActiveCollectionUpdated = new ActiveCollectionUpdatedEvent();
        public ActiveCollectionNameChangedEvent OnActiveCollectionNameChanged = new ActiveCollectionNameChangedEvent();

        private AudioMateController _c;
        private UIManager _ui;
        private readonly List<AudioMateClipCollection> _collections = new List<AudioMateClipCollection>();

        public IEnumerable<AudioMateClipCollection> CollectionsWithProvisionalClips =>
            _collections.FindAll(x => x.GetProvisionalClips().Count > 0);

        private readonly Dictionary<string,CollectionActions> _collectionActions = new Dictionary<string, CollectionActions>();

        public JSONStorableStringChooser CollectionsJSON;

        public AudioMateClipCollection ActiveCollection { get; private set; }

        public int Count => _collections.Count;

        public IEnumerable<string> Names => Enumerable.ToList(_collections.Select(collection => collection.Name));
        public bool isInitialized;
        public bool isInDefaultState = true;

        #region Initialization
        public void Init(AudioMateController controller)
        {
            if ((UnityEngine.Object) controller == (UnityEngine.Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: Controller is null.");
                return;
            }
            _c = controller;
            if (isInitialized)
            {
                Log("Collection manager is already initialized.");
                return;
            }
            Log("### Init ###");
            if ((UnityEngine.Object) _c.ui == (UnityEngine.Object) null)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(Init)}: UI is not initialized.");
                return;
            }
            CollectionsJSON = _c.CollectionsJSON;
            CollectionsJSON.setCallbackFunction = SelectActiveCollection;
            SelectActiveCollection();
            isInitialized = true;
        }

        private void AddDefaultCollection()
        {
            if (_collections == null) return;

            var newCollection = AddCollection();
            SelectActiveCollection(newCollection);
            isInDefaultState = true;
        }
        #endregion

        #region Helper
        private string GetAvailableDefaultName()
        {
            var newName = "Untitled";
            var i = 2;
            while (Names.Contains(newName))
            {
                newName = "Untitled" + i;
                i++;
            }

            return newName;
        }

        public void SyncCollectionNames()
        {
            var choices = new List<string>();
            choices.AddRange(Names);
            CollectionsJSON.choices = choices;
        }
        #endregion

        #region Collection Management

        public AudioMateClipCollection AddCollection(string collectionName = null)
        {
            if (collectionName == null)
            {
                collectionName = GetAvailableDefaultName();
            }
            var newCollection = new AudioMateClipCollection(collectionName);
            return AddCollection(newCollection);
        }

        private AudioMateClipCollection AddCollection(AudioMateClipCollection collection)
        {
            if (collection == null || string.IsNullOrEmpty(collection.Name)) return null;
            if (collection.ReceiverAtom == null) collection.ReceiverAtom = _c.containingAtom.uid;
            if (collection.ReceiverNode == null) collection.ReceiverNode = AudioMateController.GuessInitialReceivingNode(_c.containingAtom);
            isInDefaultState = false;
            _collections.Add(collection);
            AddNewCollectionActions(collection.Name);
            if (!CollectionsJSON.choices.Contains(collection.Name))
            {
                CollectionsJSON.choices.Add(collection.Name);
            }
            SelectActiveCollection(collection);
            return collection;
        }

        public bool RemoveCollection(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName)) return false;
            var collectionToRemove = _collections.Find(x => x.Name == collectionName);
            return collectionToRemove != null && RemoveCollection(collectionToRemove);
        }

        private bool RemoveCollection(AudioMateClipCollection collection)
        {
            var collectionName = collection?.Name;
            if (string.IsNullOrEmpty(collectionName)) return false;
            CollectionsJSON.choices.Remove(collectionName);
            RemoveCollectionActions(collectionName);
            var result = _collections.Remove(collection);
            SyncCollectionNames();
            return result;
        }

        public bool RemoveActiveCollection()
        {
            var collectionIndex = _collections.IndexOf(ActiveCollection);
            var result = RemoveCollection(ActiveCollection);
            if (_collections.Count == 0)
            {
                AddDefaultCollection();
            }
            else
            {
                collectionIndex--;
                SelectActiveCollection(collectionIndex == -1 ? _collections.Last() : _collections[collectionIndex]);
            }
            return result;
        }

        public void SelectActiveCollection(AudioMateClipCollection collection = null)
        {
            while (true)
            {
                if (collection != null)
                {
                    var previous = ActiveCollection;
                    ActiveCollection = collection;
                    ActiveCollection.SyncAudioReceiver();
                    if (CollectionsJSON.val != ActiveCollection.Name) CollectionsJSON.val = ActiveCollection.Name;
                    OnActiveCollectionSelected.Invoke(new ActiveCollectionSelectedEventArgs { Before = previous, After = ActiveCollection });
                    return;
                }

                if (_collections.Count > 0)
                {
                    collection = _collections.First();
                    continue;
                }
                else
                {
                    AddDefaultCollection();
                }

                break;
            }
        }

        public void SelectActiveCollection(string collectionName)
        {
            if ((UnityEngine.Object) _c == (UnityEngine.Object) null || string.IsNullOrEmpty(collectionName)) return;
            var collection = _collections.Find(x => x.Name == collectionName);
            if (collection == null) return;
            SelectActiveCollection(collection);
        }

        public void Enable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Enable();
        }

        public void Enable()
        {
            ActiveCollection?.Enable();
        }

        public void Disable(string collectionName)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            collection?.Disable();
        }

        public void Disable()
        {
            ActiveCollection?.Disable();
        }

        public void EnableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Enable();
            }
        }

        public void DisableAll()
        {
            foreach (var collection in _collections)
            {
                collection.Disable();
            }
        }

        public void SetActiveCollectionName(string newCollectionName)
        {
            if (string.IsNullOrEmpty(newCollectionName)) return;
            if (ActiveCollection == null) return;
            isInDefaultState = false;
            var oldCollectionName = ActiveCollection.Name;
            ActiveCollection.Name = newCollectionName;
            SyncCollectionNames();
            CollectionsJSON.val = newCollectionName;
            UpdateCollectionActionNames(oldCollectionName, newCollectionName);

            OnActiveCollectionNameChanged.Invoke(new ActiveCollectionNameChangedEventArgs
            {
                Before = oldCollectionName,
                After = newCollectionName
            });
        }

        /**
         * Adds a clip to the active collections
         */
        public void AddClipToActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            ActiveCollection?.Add(clip);
            isInDefaultState = false;
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds a range of clips to the active collections
         */
        public void AddClipRangeToActiveCollection(int fromIndex, int toIndex, bool noEvent = false)
        {
            if (ActiveCollection == null) return;
            isInDefaultState = false;
            for (var i = fromIndex; i <= toIndex; i++)
            {
                var clip = _c.audioMateClipLibrary.Clips.ElementAtOrDefault(i);
                if (clip == null) continue;
                ActiveCollection.Add(clip);
                _c.audioMateClipLibrary.SetCursorIndex(i);
            }
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes a clip from a collection
         */
        public bool RemoveClipFromActiveCollection(AudioMateClip clip, bool noEvent = false)
        {
            if (ActiveCollection == null || !ActiveCollection.Remove(clip)) return false;
            if (!noEvent) OnActiveCollectionUpdated.Invoke();
            return true;

        }

        /**
         * Adds all available clips to the active collections
         */
        public void AddAllClipsToActiveCollection()
        {
            foreach (var clip in _c.audioMateClipLibrary.Clips)
            {
                AddClipToActiveCollection(clip, true);
            }

            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds 10 clip to the active collections, starting from the cursor
         */
        public void Add10ClipsToActiveCollection()
        {
            AddClipRangeToActiveCollection(_c.audioMateClipLibrary.Cursor.Index, _c.audioMateClipLibrary.Cursor.Index + 10, true);
            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Adds 20 clip to the active collections, starting from the cursor
         */
        public void Add20ClipsToActiveCollection()
        {
            AddClipRangeToActiveCollection(_c.audioMateClipLibrary.Cursor.Index, _c.audioMateClipLibrary.Cursor.Index + 20, true);
            OnActiveCollectionUpdated.Invoke();
        }

        /**
         * Removes all clips from the active collections
         */
        public void ClearActiveCollection()
        {
            ActiveCollection?.Clear();
            OnActiveCollectionUpdated.Invoke();
        }
        #endregion

        #region Clip Actions
        /**
         * Plays a random clip from a specific Collection
         */
        public void PlayRandomClipInCollection(string collectionName, bool queue = false)
        {
            var collection = _collections.Find(x => x.Name == collectionName);
            if (queue)
            {
                collection?.QueueRandomClip();
            }
            else
            {
                collection?.PlayRandomClip();
            }
        }

        /**
         * Queues a random clip from a specific Collection
         */
        public void QueueRandomClipInCollection(string collectionName)
        {
            PlayRandomClipInCollection(collectionName, true);
        }

        public void PlayRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.PlayRandomClip();
            }
            else
            {
                PlayRandomClipInCollection(collectionName);
            }
        }

        public void QueueRandomClipAction(string collectionName = null)
        {
            if (collectionName == null)
            {
                ActiveCollection?.QueueRandomClip();
            }
            else
            {
                QueueRandomClipInCollection(collectionName);
            }
        }

        public List<AudioMateClip> UsedClips(string collectionFilter = null)
        {
            var result = new List<AudioMateClip>();

            if (collectionFilter == null)
            {
                foreach (var collection in _collections)
                {
                    result.AddRange(collection.Clips);
                }
            }
            else
            {
                var collection = _collections.Find(x => x.Name == collectionFilter);
                if (collection == null) return result;
                result = collection.Clips;
            }

            return result;
        }
        #endregion

        /**
         * Creates all needed Actions for a new collection
         */
        public void AddNewCollectionActions(string collectionName)
        {
            if (string.IsNullOrEmpty(collectionName))
            {
                Log($"{nameof(AddNewCollectionActions)}: collectionName is empty");
                return;
            }
            if (_collectionActions == null) return;
            if (_collectionActions.ContainsKey(collectionName)) return;
            var newActions = new CollectionActions
            {
                PlayAction = new JSONStorableAction(GetPlayRandomClipActionName(collectionName),
                    () => PlayRandomClipAction(collectionName)),
                QueueAction = new JSONStorableAction(GetQueueRandomClipActionName(collectionName),
                    () => QueueRandomClipAction(collectionName))
            };

            _c.RegisterAction(newActions.PlayAction);
            _c.RegisterAction(newActions.QueueAction);
            _collectionActions.Add(collectionName, newActions);
        }

        private void UpdateCollectionActionNames(string oldCollectionName, string newCollectionName)
        {
            if (string.IsNullOrEmpty(oldCollectionName) || string.IsNullOrEmpty(newCollectionName))
            {
                Log($"{nameof(AddNewCollectionActions)}: one or both input values are empty");
                return;
            }
            if (_collectionActions == null) return;
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(oldCollectionName, out actions)) return;
            if (actions != null)
            {
                RemoveCollectionActions(oldCollectionName);
            }
            AddNewCollectionActions(newCollectionName);
        }

        private void RemoveCollectionActions(string collectionName)
        {
            CollectionActions actions;
            if (!_collectionActions.TryGetValue(collectionName, out actions)) return;
            _c.DeregisterAction(actions.PlayAction);
            _c.DeregisterAction(actions.QueueAction);
            _collectionActions.Remove(collectionName);
            actions.PlayAction = null;
            actions.QueueAction = null;
        }

        public static string GetPlayRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.PlayRandomClipInCollectionAction}{collectionName}");
        }

        public static string GetQueueRandomClipActionName(string collectionName)
        {
            return Tools.TrimAll($"{Storables.QueueRandomClipInCollectionAction}{collectionName}");
        }

        #region Callbacks

        public void OnAtomRename(string oldID, string newID)
        {
            foreach (var collection in _collections.Where(collection => collection.ReceiverAtom == oldID))
            {
                collection.ReceiverAtom = newID;
                collection.SyncAudioReceiver();
            }
        }
        #endregion

        #region Serializer

        public void Serialize(JSONClass json, string id, string collectionName = null,  bool resetAudioSourceIds = false)
        {
            if (json == null) json = new JSONClass {{id, null}};
            if (json[id] == null) json.Add(id, null);
            json[id] = GetCollectionsJSON(collectionName, resetAudioSourceIds);
        }

        private JSONClass GetCollectionsJSON(string collectionName = null, bool resetAudioSourceIds = false)
        {
            var json = new JSONClass();
            var receiverAtomId = _c.containingAtom.uid;
            try
            {
                var query = _collections.Where(collection => collection != null && collection.Count != 0);
                if (collectionName != null) query = query.Where(collection => collection.Name == collectionName);
                foreach (var collection in query)
                {
                    json[collection.Name] = collection.ToJSON();
                    if (resetAudioSourceIds && json[collection.Name]["receiverAtom"].Value == receiverAtomId)
                    {
                        json[collection.Name]["receiverAtom"] = "Person";
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError($"AudioMate.{nameof(CollectionManager)}.{nameof(GetCollectionsJSON)}: {e}");
            }
            return json;
        }

        public void Parse(JSONNode json, bool merge = false)
        {
            DisableAll();
            if (json == null || !json.Childs.Any()) return;
            //if (json[Storables.ClipCollections] != null) json = json[Storables.ClipCollections];
            try
            {
                if (merge == false) _collections.Clear();
                foreach (var collectionJSON in json.Childs.ToList())
                {
                    var mergeTarget = _collections.Find(x => x.Name == collectionJSON["name"]);
                    var collection = mergeTarget ?? new AudioMateClipCollection(collectionJSON["name"]);
                    if (collection.Parse(collectionJSON, merge) && mergeTarget == null)
                    {
                        AddCollection(collection);
                    }
                }

                EnableAll();
                SelectActiveCollection(_collections.First());
                _c.needsProvisionalClipProcessing = true;
            }
            catch (Exception e)
            {
                SuperController.LogError(
                    $"AudioMate.{nameof(CollectionManager)}.{nameof(Parse)}: {e}");
            }
        }

        #endregion

        private void Log(string message)
        {
            if ((UnityEngine.Object) _c == (UnityEngine.Object) null) return;
            _c.Log($"CollectionManager: {message}");
        }

    }
}
