﻿using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using UnityEngine;

namespace AudioMate
{
    public class ScreenCollection : Screen
    {
        public new static readonly ScreenDefinition Definition = new ScreenDefinition(
            "Collection",
            "Playback / Triggers",
            typeof(ScreenCollection));

        private JSONStorableBool _toggleOnlyIfClearJSON;
        private JSONStorableBool _toggleAlwaysQueueJSON;
        private JSONStorableBool _toggleShuffleJSON;
        private JSONStorableFloat _sliderPlayChanceJSON;
        private bool _needsRefresh;

        private List<HorizontalButtonGroup> _triggerActionsUI = new List<HorizontalButtonGroup>();

        public override void Init(AudioMateController controller)
        {
            base.Init(controller);

            InitPlaybackOptionsUI();
            InitTriggerActionsUI();
        }

        private void InitTriggerActionsUI()
        {
            _ui.CreateHeader("Trigger Actions", 2);
            _ui.CreatePopup(_c.TriggerColliderChooserJSON, true);
            _c.TriggerColliderChooserJSON.setCallbackFunction = val =>
            {
                _needsRefresh = true;
            };
            var btnGroup = _ui.gameObject.AddComponent<HorizontalButtonGroup>();
            btnGroup.CreateButton("Add Start Action", Styles.Button, () =>
            {
                _c.OnAddTriggerButtonClicked();
                _needsRefresh = true;
            });
            btnGroup.CreateButton("Add End Action", Styles.Button, () =>
            {
                _c.OnAddTriggerButtonClicked(TriggerManager.EndTriggerActions);
                _needsRefresh = true;
            });
            _needsRefresh = true;
        }

        private void InitPlaybackOptionsUI()
        {
            _ui.CreateHeader("Playback", 2);
            _sliderPlayChanceJSON = new JSONStorableFloat("Play Chance", 1f, (float val) =>
            {
                _c.collections.ActiveCollection.PlayChance = val;
            }, 0f, 1f);
            _ui.CreateSlider(_sliderPlayChanceJSON);
            _toggleOnlyIfClearJSON = new JSONStorableBool("Play Only If Clear", false, (bool val) =>
            {
                _c.collections.ActiveCollection.OnlyIfClear = val;
            });
            _ui.CreateToggle(_toggleOnlyIfClearJSON);

            _toggleAlwaysQueueJSON = new JSONStorableBool("Always Queue", false, (bool val) =>
            {
                _c.collections.ActiveCollection.AlwaysQueue = val;
            });
            _ui.CreateToggle(_toggleAlwaysQueueJSON);

            _toggleShuffleJSON = new JSONStorableBool("Shuffle Mode", true, (bool val) =>
            {
                _c.collections.ActiveCollection.ShuffleMode = val;
            });
            _ui.CreateToggle(_toggleShuffleJSON);
        }

        private void RefreshTriggerActionsUI()
        {
            var activeCollection = _c.collections.ActiveCollection;
            var triggerCollection = _c.Triggers.AllCollectionTriggerActions(activeCollection.Name);
            foreach (var uiElem in _triggerActionsUI)
            {
                uiElem.Destroy();
            }

            var triggerID = _c.TriggerColliderChooserJSON.val;
            var triggerActions = triggerCollection[triggerID];
            if (triggerActions[TriggerManager.StartTriggerActions].Childs.Any()) {
                foreach (JSONClass tA in triggerActions[TriggerManager.StartTriggerActions].AsArray)
                {
                    CreateTriggerActionUI(triggerID, tA, TriggerManager.StartTriggerActions);
                }
            }
            if (triggerActions[TriggerManager.EndTriggerActions].Childs.Any())
            {
                foreach (JSONClass tA in triggerActions[TriggerManager.EndTriggerActions].AsArray)
                {
                    CreateTriggerActionUI(triggerID, tA, TriggerManager.EndTriggerActions);
                }
            }
        }

        private void CreateTriggerActionUI(string triggerID, JSONClass triggerActionJSON, string triggerActionType)
        {
            var buttonGroup = _ui.CreateButtonGroup();
            buttonGroup.CreateLabel(triggerActionType == TriggerManager.StartTriggerActions ? "Start Trigger Action" : "End Trigger Action", Styles.Button, true, 26);
            buttonGroup.CreateButton("Delete", Styles.Danger, () =>
            {
                _needsRefresh = true;
                _c.Triggers.RemoveTriggerAction(triggerID, triggerActionJSON["name"].ToString(), triggerActionType);
            }, false, 50f, 0);

            _triggerActionsUI.Add(buttonGroup);
        }

        protected override void OnActiveCollectionSelected(CollectionManager.ActiveCollectionSelectedEventArgs args)
        {
            base.OnActiveCollectionSelected(args);
            var activeCollection = args.After;
            _toggleOnlyIfClearJSON.val = activeCollection.OnlyIfClear;
            _toggleAlwaysQueueJSON.val = activeCollection.AlwaysQueue;
            _toggleShuffleJSON.val = activeCollection.ShuffleMode;
            _sliderPlayChanceJSON.val = activeCollection.PlayChance;
            _needsRefresh = true;
        }

        private void FixedUpdate()
        {
            if (!_needsRefresh) return;
            _needsRefresh = false;
            RefreshTriggerActionsUI();
        }
    }
}
