using System;
using System.Collections;
using System.Collections.Generic;
using AssetBundles;
using Leap;
using UnityEngine;
using UnityEngine.UI;
using Image = UnityEngine.UI.Image;

namespace AudioMate
{
    /// <summary>
    /// Base code shamelessly ripped from the incredible Timeline plugin (https://github.com/acidbubbles/vam-timeline)
    /// Shout outs to Acidbubbles!
    /// </summary>
    public class UIFactory : MonoBehaviour
    {
        private static RectTransform _scrollbarPrefab;
        public static RectTransform ButtonPrefab;
        public static readonly Font BaseFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
        private GameObject _currentContainer;

        public static IEnumerator LoadUIAssets()
        {
            foreach (var x in LoadUIAsset("z_ui2", "DynamicTextField", prefab => _scrollbarPrefab = prefab.GetComponentInChildren<ScrollRect>().verticalScrollbar.gameObject.GetComponent<RectTransform>())) yield return x;
            foreach (var x in LoadUIAsset("z_ui2", "DynamicButton", prefab => ButtonPrefab = prefab)) yield return x;
        }

        private static IEnumerable LoadUIAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            var request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            var go = request.GetAsset<GameObject>();
            if (go == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            var prefab = go.GetComponent<RectTransform>();
            if (prefab == null) throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }

        public static ScrollRect CreateScrollRect(GameObject gameObject)
        {
            var scrollView = CreateScrollView(gameObject);
            var viewport = CreateViewport(scrollView);
            var content = CreateContent(viewport);
            var scrollbar = CreateScrollbar(scrollView);
            var scrollRect = scrollView.GetComponent<ScrollRect>();
            scrollRect.viewport = viewport.GetComponent<RectTransform>();
            scrollRect.content = content.GetComponent<RectTransform>();
            scrollRect.verticalScrollbar = scrollbar;
            scrollRect.verticalScrollbarVisibility = ScrollRect.ScrollbarVisibility.AutoHideAndExpandViewport;
            return scrollRect; //content.GetComponent<RectTransform>();
        }

        private static GameObject CreateScrollView(GameObject parent)
        {
            var go = new GameObject("Scroll View");
            go.transform.SetParent(parent.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();

            var scroll = go.AddComponent<ScrollRect>();
            scroll.horizontal = false;

            return go;
        }

        private static Scrollbar CreateScrollbar(GameObject scrollView)
        {
            var vs = Instantiate(_scrollbarPrefab, scrollView.transform, false);
            return vs.GetComponent<Scrollbar>();
        }

        private static GameObject CreateViewport(GameObject scrollView)
        {
            var go = new GameObject("Viewport");
            go.transform.SetParent(scrollView.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchParent();
            rect.pivot = new Vector2(0, 1);

            var image = go.AddComponent<Image>();
            image.raycastTarget = true;

            var mask = go.AddComponent<Mask>();
            mask.showMaskGraphic = false;

            return go;
        }

        private static GameObject CreateContent(GameObject viewport)
        {
            var go = new GameObject("Content");
            go.transform.SetParent(viewport.transform, false);

            var rect = go.AddComponent<RectTransform>();
            rect.StretchTop();
            rect.pivot = new Vector2(0, 1);

            var layout = go.AddComponent<VerticalLayoutGroup>();
            layout.spacing = 10f;
            layout.childForceExpandWidth = true;
            layout.childForceExpandHeight = false;
            layout.childControlWidth = true;
            layout.childControlHeight = true;
            layout.childAlignment = TextAnchor.UpperLeft;

            var fit = go.AddComponent<ContentSizeFitter>();
            fit.horizontalFit = ContentSizeFitter.FitMode.Unconstrained;
            fit.verticalFit = ContentSizeFitter.FitMode.PreferredSize;

            return go;
        }

        public AudioMateController c;
        private readonly List<JSONStorableParam> _storables = new List<JSONStorableParam>();

        private static string PrepareLabelStyle(string label)
        {
            //return label.ToUpper();
            return label;
        }

        public UIDynamic CreateSpacer()
        {
            var ui = Instantiate(c.Manager.configurableSpacerPrefab).GetComponent<UIDynamic>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.height = 20f;
            return ui;
        }

        public Text CreateText(string content = "", int fontSize = 32, TextAnchor alignment = TextAnchor.UpperLeft)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);
            var rect = go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);

            var layout = go.AddComponent<LayoutElement>();
            layout.flexibleHeight = 1;
            layout.flexibleWidth = 1;

            var text = rect.gameObject.AddComponent<Text>();
            text.text = content;
            text.font = BaseFont;
            text.fontSize = fontSize;
            text.color = Color.black;
            text.alignment = alignment;

            return text;
        }

        public UIDynamicSlider CreateSlider(JSONStorableFloat jsf, float scale = 1.0f, bool showQuickButtons = false)
        {
            RegisterStorable(jsf);
            var prefab = c.Manager.configurableSliderPrefab;
            prefab.localScale = new Vector3(scale, scale);
            var ui = Instantiate(c.Manager.configurableSliderPrefab).GetComponent<UIDynamicSlider>();
            ui.gameObject.transform.SetParent(transform, false);
            var rectTransform = ui.gameObject.GetComponent<RectTransform>();
            rectTransform.pivot = new Vector2(0, 0.5f);
            var image = rectTransform.GetChild(0).gameObject.GetComponent<Image>();
            image.color = Styles.ButtonBg;
            ui.labelText.color = Styles.ButtonText;
            //_c.DebugObject(ui.gameObject);
            ui.Configure(jsf.name, jsf.min, jsf.max, jsf.val, jsf.constrained, "F2", showQuickButtons, !jsf.constrained);
            jsf.slider = ui.slider;
            return ui;
        }

        public UIDynamicButton CreateButton(string label, int style = Styles.Button, bool addOutline = false)
        {
            var ui = Instantiate(c.Manager.configurableButtonPrefab).GetComponent<UIDynamicButton>();
            ui.buttonColor = Styles.Bg(style);
            ui.textColor = Styles.Text(style);
            ui.buttonText.fontSize = 22;
            var outline = ui.button.gameObject.AddComponent<Outline>().GetComponent<Outline>();
            outline.effectColor = addOutline ? new Color(0f,0f,0f,0.5f) : new Color(0f,0f,0f,0f);
            outline.effectDistance = new Vector2(2f,2f);
            outline.enabled = addOutline;
            ui.gameObject.transform.SetParent(transform, false);
            ui.label = PrepareLabelStyle(label.ToUpper());
            return ui;
        }

        public HorizontalButtonGroup CreateButtonGroup()
        {
            return gameObject.AddComponent<HorizontalButtonGroup>();
        }

        public UIDynamicToggle CreateToggle(JSONStorableBool jsb, int style = Styles.Button)
        {
            RegisterStorable(jsb);
            var ui = Instantiate(c.Manager.configurableTogglePrefab).GetComponent<UIDynamicToggle>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.backgroundColor = Styles.Bg(style);
            ui.textColor = Styles.Text(style);
            ui.label = jsb.name;
            ui.toggle.graphic.color = Styles.Bg(style);
            jsb.toggle = ui.toggle;
            return ui;
        }

        public UIDynamicTextField CreateTextField(JSONStorableString jss)
        {
            RegisterStorable(jss);
            var ui = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.UItext.fontSize = 28;
            ui.backgroundColor = Styles.EmbeddedBg;
            ui.textColor = Styles.EmbeddedText;
            ui.backgroundImage.fillMethod = Image.FillMethod.Radial180;
            jss.dynamicText = ui;
            return ui;
        }

        public UIDynamicPopup CreatePopup(JSONStorableStringChooser jsc, bool filterable)
        {
            RegisterStorable(jsc);
            Transform prefab;
            #if(VAM_GT_1_20)
            if(filterable)
                prefab = c.manager.configurableFilterablePopupPrefab;
            else
            #endif
                prefab = c.Manager.configurableScrollablePopupPrefab;
            var ui = Instantiate(prefab).GetComponent<UIDynamicPopup>();
            ui.gameObject.transform.SetParent(transform, false);
            ui.labelTextColor = Styles.ButtonText;
            ui.popup.backgroundImage.color = Styles.ButtonBg;
            ui.label = jsc.name;
            ui.popup.label = jsc.val;
            jsc.popup = ui.popup;
            return ui;
        }

        public UIDynamicTextField CreateTextInput(
            JSONStorableString jss,
            bool hasLabel = false,
            InputField.ContentType contentType = InputField.ContentType.Alphanumeric)
        {
            RegisterStorable(jss);

            var go = new GameObject();
            go.transform.SetParent(transform, false);
            {
                var rect = go.AddComponent<RectTransform>();
                rect.pivot = new Vector2(0, 1);

                var layout = go.AddComponent<LayoutElement>();
                layout.preferredHeight = 25f;
                layout.flexibleHeight = 0;
                layout.flexibleWidth = 1f;
            }

            var textfield = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            textfield.gameObject.transform.SetParent(go.transform, false);
            {
                jss.dynamicText = textfield;

                textfield.backgroundColor = Styles.InputBg;
                textfield.textColor = Styles.InputText;
                var input = textfield.gameObject.AddComponent<InputField>();
                textfield.UItext.resizeTextForBestFit = true;
                textfield.UItext.fontSize = 42;
                textfield.UItext.alignment = TextAnchor.LowerLeft;
                textfield.UItext.alignByGeometry = true;
                input.textComponent = textfield.UItext;
                input.contentType = contentType;
                jss.inputField = input;
                input.lineType = InputField.LineType.SingleLine;
                //input.textComponent.alignment = TextAnchor.LowerLeft;
                var rect = textfield.GetComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 0);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 0);
                rect.anchoredPosition = new Vector2(0, 0);
                rect.sizeDelta = new Vector2(0, 25f);
                var layout = textfield.GetComponent<LayoutElement>();
                Destroy(layout);
            }
            if (!hasLabel) return textfield;
            var title = new GameObject();
            title.transform.SetParent(go.transform, false);
            {
                var rect = title.AddComponent<RectTransform>();
                rect.anchorMin = new Vector2(0, 0);
                rect.anchorMax = new Vector2(1, 1);
                rect.pivot = new Vector2(0, 1);
                rect.anchoredPosition = new Vector2(0, 1f);
                rect.sizeDelta = new Vector2(0, 30f);

                var text = title.AddComponent<Text>();
                text.font = textfield.UItext.font;
                text.text = jss.name;
                text.fontSize = 24;
                text.color = Styles.ButtonBg;
            }
            return textfield;
        }

        public UIDynamicTextField CreateSimpleTextInput(
            JSONStorableString jss,
            int fontSize = 42,
            InputField.ContentType contentType = InputField.ContentType.Alphanumeric)
        {
            RegisterStorable(jss);
            var go = new GameObject();


                go.transform.SetParent(transform, false);

                var rect = go.AddComponent<RectTransform>();
                rect.pivot = new Vector2(0, 1);
                var blayout = go.gameObject.AddComponent<LayoutElement>();
                blayout.preferredHeight = fontSize * 1.2f;
                blayout.flexibleHeight = 0;
                blayout.flexibleWidth = 1;


        var textfield = Instantiate(c.Manager.configurableTextFieldPrefab).GetComponent<UIDynamicTextField>();
            textfield.gameObject.transform.SetParent(go.transform, false);

                jss.dynamicText = textfield;

                textfield.backgroundColor = Color.clear;
                textfield.textColor = Color.white;
                var input = textfield.gameObject.AddComponent<InputField>();
                textfield.UItext.resizeTextForBestFit = true;
                textfield.UItext.fontSize = fontSize;
                textfield.UItext.fontStyle = FontStyle.Bold;
                //var layout = textfield.UItext.gameObject.AddComponent<LayoutElement>();

                textfield.UItext.alignment = TextAnchor.UpperLeft;

                input.textComponent = textfield.UItext;
                input.contentType = contentType;
                jss.inputField = input;
                input.lineType = InputField.LineType.SingleLine;
                input.textComponent.alignment = TextAnchor.LowerLeft;
                var rect2 = textfield.GetComponent<RectTransform>();
                rect2.anchorMin = new Vector2(0, 0);
                rect2.anchorMax = new Vector2(1, 1);
                rect2.pivot = new Vector2(0, 0);
                rect2.anchoredPosition = new Vector2(0, 20f);
                rect2.sizeDelta = new Vector2(0, 0);
                var layout = textfield.GetComponent<LayoutElement>();

                Destroy(layout);

            return textfield;
        }

        public Text CreateHeader(string headline, int level = 1, int fixedHeight = 0, TextAnchor alignment = TextAnchor.LowerLeft)
        {
            var go = new GameObject();
            go.transform.SetParent(transform, false);
            var rect = go.AddComponent<RectTransform>();
            rect.pivot = new Vector2(0, 1);

            var layout = go.AddComponent<LayoutElement>();

            var text = rect.gameObject.AddComponent<Text>();
            text.text = PrepareLabelStyle(headline);
            text.font = BaseFont;
            text.alignment = alignment;
            if (level > 3) level = 3;
            switch (level)
            {
                case 3:
                    text.fontSize = 24;
                    text.fontStyle = FontStyle.Bold;
                    text.color = Color.black;
                    break;
                case 2:
                    text.fontSize = 32;
                    text.fontStyle = FontStyle.Bold;
                    text.color = Color.black;
                    break;
                default:
                    text.fontSize = 42;
                    text.fontStyle = FontStyle.Bold;
                    text.color = Color.black;
                    break;
            }

            layout.preferredHeight = fixedHeight == 0 ? text.fontSize * 2 : fixedHeight;
            layout.flexibleHeight = fixedHeight == 0 ? 1 : 0;
            layout.flexibleWidth = 1;
            layout.minHeight = fixedHeight;

            return text;
        }

         public void ClearOverlay()
        {
            Destroy(_currentContainer);
            _currentContainer = null;
        }

         public void CreateOverlay(string message, OverlaySettings settings)
        {
            if (_currentContainer != null) throw new InvalidOperationException("Only one overlay can be shown at a time.");

            var container = new GameObject();
            container.transform.SetParent(transform.parent.parent, false);

            {
                var rect = container.AddComponent<RectTransform>();
                rect.anchorMin = new Vector2(0f, 0f);
                rect.anchorMax = new Vector2(1f, 1f);

                var bg = container.AddComponent<Image>();
                bg.color = Styles.OverlayBg;
                bg.preserveAspect = false;
            }

            var group = new GameObject();
            group.transform.SetParent(container.transform, false);
            {
                var rect = group.AddComponent<RectTransform>();
                rect.anchorMin = new Vector2(0f, 0.5f);
                rect.anchorMax = new Vector2(1f, 0.5f);
                rect.sizeDelta = new Vector2(0f, 200f);
            }
            var layout = group.AddComponent<VerticalLayoutGroup>();
            layout.childControlHeight = true;
            layout.childForceExpandHeight = true;
            layout.childControlWidth = true;
            layout.childForceExpandWidth = true;

            layout.spacing = 10f;
            layout.padding = new RectOffset(50, 50, 50, 50);


                var text = CreateHeader(message, 2);
                text.color = Styles.OverlayText;
                text.alignment = TextAnchor.MiddleCenter;
                text.transform.SetParent(group.transform, false);


            // TODO implement argument buttons

            var mainButtons = settings.MainButtons;
            foreach (var button in mainButtons)
            {
                var btn = CreateButton(button.Label);
                btn.transform.SetParent(text.transform, false);
                btn.buttonImage.fillCenter = false;

                btn.buttonColor = Styles.Text(button.Style);
                btn.textColor = Styles.Text(button.Style);

                btn.button.onClick.AddListener(() =>
                {
                    _currentContainer = null;
                    Destroy(container);
                    button.Callback();
                });
            }
            _currentContainer = container;
        }

         public void OnDestroy()
        {
            var clone = new JSONStorableParam[_storables.Count];
            _storables.CopyTo(clone);
            _storables.Clear();
            foreach (var component in clone)
            {
                if (component == null) continue;
                if (component is JSONStorableStringChooser)
                    RemovePopup((JSONStorableStringChooser)component);
                else if (component is JSONStorableFloat)
                    RemoveSlider((JSONStorableFloat)component);
                else if (component is JSONStorableString)
                    RemoveTextField((JSONStorableString)component);
                else if (component is JSONStorableBool)
                    RemoveToggle((JSONStorableBool)component);
                else
                    SuperController.LogError($"AudioMate: Cannot remove component {component}");
            }
        }

        public void RemovePopup(JSONStorableStringChooser jsc, UIDynamicPopup component = null)
        {
            if (jsc.popup != null) { jsc.popup = null; _storables.Remove(jsc); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveSlider(JSONStorableFloat jsf, UIDynamicSlider component = null)
        {
            if (jsf.slider != null) { jsf.slider = null; _storables.Remove(jsf); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveTextField(JSONStorableString jss, UIDynamicTextField component = null)
        {
            if (jss.dynamicText != null) { jss.dynamicText = null; _storables.Remove(jss); }
            if (component != null) Destroy(component.gameObject);
        }

        public void RemoveToggle(JSONStorableBool jsb, UIDynamicToggle component = null)
        {
            if (jsb.toggle != null) { jsb.toggle = null; _storables.Remove(jsb); }
            if (component != null) Destroy(component.gameObject);
        }

        private T RegisterStorable<T>(T v)
            where T : JSONStorableParam
        {
            _storables.Add(v);
            ValidateStorableFreeToBind(v);
            return v;
        }

        private void ValidateStorableFreeToBind(JSONStorableParam v)
        {
            if (v is JSONStorableStringChooser)
            {
                if (((JSONStorableStringChooser)v).popup != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableFloat)
            {
                if (((JSONStorableFloat)v).slider != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableString)
            {
                if (((JSONStorableString)v).inputField != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
            else if (v is JSONStorableBool)
            {
                if (((JSONStorableBool)v).toggle != null)
                    SuperController.LogError($"Storable {v.name} of atom {c.containingAtom.name} was not correctly unregistered.");
            }
        }
    }
}
